/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 1994-2007, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

------------------------------------------------------------------------------

dwatch.c watches a directory for files that are created, modified, or
deleted.

$Revision: 0.0 $
$Date: 2007/09/10 19:36:00 $
$Id: dwatch.c,v 1.7 2007/09/10 19:36:00 john Exp $
$Log: dwatch.c,v $
Revision 0.0  2007/09/10 19:36:00  john
Initial revision

*/

#define _GNU_SOURCE /* required for fcntl(2) F_NOTIFY defines */

#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

static char rcsid[] = "$Id: dwatch.c,v 0.0 2007/09/10 19:36:00 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 1994-2007, John Conover, All Rights Reserved"; /* the copyright banner */

static void handler (int sig, siginfo_t *si, void *data); /* signal handler to retreive data */
static void print_message (int retval); /* print any error messages */

static const char *help_message[] = /* help message index array */
{
    "\n",
    "Watch a directory for changes\n",
    "Usage: dwatch [-d directory] [-v]\n",
    "    -d directory, directory to watch, (.)\n",
    "    -v, print the program's version information\n"
};

static const char *error_message[] = /* error message index array */
{
    "No error\n",
    "Error in program argument(s)\n",
    "Error opening directory\n",
    "Error initializing signal handler\n",
    "Error installing signal handler\n",
    "Error manipulating file descriptor\n"
};

#define NOERROR 0 /* error values, one for each index in the error message array */
#define EARGS 1
#define EOPEN 2
#define ESIGINT 3
#define ESIG 4
#define EFCNTL 5

static volatile int event_fd; /* the file/directory descriptor */
static volatile int event_sig; /* the signal */
static volatile void *event_data; /* save any data */

static char default_dir[] = "."; /* default directory to watch, CWD */

int main (int argc, char *argv[])
{
    char *dir = default_dir; /* default directory is CWD */

    int retval = NOERROR, /* return value, assume no error */
        fd, /* file/directory descriptor */
        c; /* command line switch */

    struct sigaction act; /* sigaction(2) structure */

    while ((c = getopt (argc, argv, "d:v")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'd': /* request for directory? */

                dir = optarg; /* yes, save the directory */
                break;

            case 'v':

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    if (retval == NOERROR) /* enough arguments? */
    {
        act.sa_sigaction = handler; /* sigaction function, handler() */
        retval = ESIGINT; /* assume error initializing signal handler */

        if (sigemptyset (&act.sa_mask) == 0) /* initialize the signal set to empty, with all signals excluded from the set */
        {
            act.sa_flags = SA_SIGINFO; /* action is to provide information */
            retval = ESIG; /* assume error installing signal handler */

            if (sigaction (SIGRTMIN + 1, &act, NULL) == 0) /* change the action of the signal */
            {
                retval = EOPEN; /* assume error opening directory */

                if ((fd = open (dir, O_RDONLY | O_DIRECTORY)) != 0) /* open the directory */
                {
                    retval = EFCNTL; /* assume error manipulating file descriptor */

                    if (fcntl (fd, F_SETSIG, SIGRTMIN + 1) != -1) /* signal when I/O becomes available on the directory */
                    {

                        if (fcntl (fd, F_NOTIFY, DN_ACCESS | DN_MODIFY | DN_CREATE | DN_RENAME | DN_DELETE | DN_ATTRIB | DN_MULTISHOT) != -1) /* notify when any file in the directory changes */
                        {
                            retval = NOERROR; /* assume no error */

                            while (1) /* while forever */
                            {
                                (void) pause (); /* wait for the signal */
                                (void) fprintf (stderr, "File(s) modified in %s\n", dir); /* notify that something changed in the directory */
                            }

                        }

                    }

                }

            }

        }

    }

    print_message (retval); /* print any error messages */
    exit (retval); /* exit with the error value */

#ifdef LINT

    return (0); /* for lint formalities */

#endif


}

/*

Signal handler to retreive data.

static void handler (int sig, siginfo_t *si, void *data);

I) Load the data into globally visable variables.

Returns nothing.

*/

static void handler (int sig, siginfo_t *si, void *data)
{
  event_fd = si->si_fd; /* save the file/directory descriptor */
  event_sig = sig; /* save the signal */
  event_data = data; /* save any data */
}

/*

Print any error messages.

static void print_message (int retval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer retval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer retval:

        1) If retval is zero, print nothing-normal/successful program
           exit.

        2) Else if retval is unity, print help.

        3) Else retval is not zero or unity, it is an error code, print
           the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int retval)

#else

static void print_message (retval)
int retval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (retval) /* which return value */
    {

        case 0: /* program ended without errors, print nothing */

            break;

        case 1: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[retval]);
            break;

    }

}
