/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 1994-2007, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

------------------------------------------------------------------------------

logwatch-exec.c is for executing a program, from a partial key
database, when the log file is appended.

When the log file, a record oriented text file with tab delimited
field(s) in each record, (specified by the -l logfilename argument,)
has record(s) appended, the record(s) are parsed, and the key field,
(specified with the -f n argument, n = 0 is the first field,) is used
for a partial key search of the first field of each record in the
database file, (specified with the -d argument); if a match is found,
the command specified in the database record is executed.

In addition, there are two internal commands:

    1) enable, (the default,) which enables external command
       execution.

    2) disable, which disables eternal command execution.

The database file is a lexically ordered, (using, perhaps, sort(1),)
text file with each record containing exactly two fields: the first is
expected to, (possibly,) match the key field from the log file; the
second is the command to execute if a match for that key is found. The
two fields are delimited by a single tab character.

The database file is opened with mmap(2), and binary searched.

$Revision: 0.0 $
$Date: 2007/09/10 19:36:00 $
$Id: logwatch-exec.c,v 1.7 2007/09/10 19:36:00 john Exp $
$Log: logwatch-exec.c,v $
Revision 0.0  2007/09/10 19:36:00  john
Initial revision

*/

#define _GNU_SOURCE /* required for fcntl(2) F_NOTIFY defines */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <libgen.h>
#include <sys/wait.h>
#include <sys/mman.h>
#include <errno.h>

static char rcsid[] = "$Id: logwatch-exec.c,v 0.0 2007/09/10 19:36:00 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 1994-2007, John Conover, All Rights Reserved"; /* the copyright banner */

#define BUFLEN BUFSIZ /* i/o buffer size */

#define TOKEN_SEPARATORS "\t\n" /* database file record field separators */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Execute a program, from a partial key database, when a log file is appended\n",
    "Usage: logwatch-exec -d dbfilename -l logfilename [-f n] [-V] [-v]\n",
    "    -d dbfilename, database filename\n",
    "    -f n, log file field number for partial key match, (0)\n",
    "    -l logfilename, log file name\n",
    "    -V, verbose mode\n",
    "    -v, print the program's version information\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error\n",
    "Error in program argument(s)\n",
    "Error opening log file\n",
    "Error closing log file\n",
    "Error seeking EOF of log file\n",
    "Error reading log file\n",
    "Error in status of database file\n",
    "Error in input buffer overflow\n",
    "Error opening database file\n",
    "Error closing database file\n",
    "Error mapping database file\n",
    "Error unmapping database file\n",
    "Error in database file\n",
    "Error initializing signal handler\n",
    "Error opening directory\n",
    "Error closing directory\n",
    "Error manipulating directory file descriptor\n",
    "Error in database buffer overflow\n",
    "Error system call\n",
    "Error allocating memory\n",
    "Error hash table already initilaized\n",
    "Error hash table mkhash () failure\n",
    "Error hash table key not found\n",
    "Error command\n"
};

#define NOERROR 0 /* error values, one for each index in the error message array */
#define EARGS 1
#define ELFILE 2
#define ELCLOSE 3
#define ELSEEK 4
#define ELREAD 5
#define EDSTAT 6
#define EOVRFL 7
#define EDOPEN 8
#define EDCLOSE 9
#define EDMAP 10
#define EDUNMAP 11
#define EDB 12
#define ESIGINT 13
#define EPOPEN 14
#define EPCLOSE 15
#define EFCNTL 16
#define EDBOVRFL 17
#define ESYS 18
#define EALLOC 19
#define EHINIT 20
#define EHMK 21
#define EHKEY 22
#define ECMD 23

static volatile int event_fd; /* the file/directory descriptor */
static volatile int event_sig; /* the signal */
static volatile void *event_data; /* save any data */

enum command /* commands */
{
    ENABLE,
    DISABLE,
    NONE
};

static const char *keyword[] = /* command interpreter keywords, must be the same number, and in same sequence as enum command */
{
    "enable",
    "disable",
    "none"
};

typedef struct hash_struct /* hash table element, the required elements are struct hash_struct *previous; and struct hash_struct *next; other appropriate elements may be added, consistent with the capabilities of the functions in HASHTABLE */
{
    struct hash_struct *previous; /* reference to next element in hash's doubly, circular linked list */
    void *hash_key; /* reference to element's key */
    int command; /* command code, one of enum keyword */
    struct hash_struct *next; /* reference to previous element in hash's doubly, circular linked list */
} HASH;

#ifdef __STDC__

typedef struct hashtable_struct /* hash table descriptor */
{
    size_t hash_size; /* size of hash table */
    HASH *table; /* reference to hash table index */
    HASH *(*mkhash) (void *key); /* reference to the function that allocates hash elements */
    int (*cmphash) (void *data, struct hash_struct *element); /* reference to the function that compares element's keys */
    void (*rmhash) (struct hash_struct *element); /* reference to the function that deallocates hash elements */
    int (*comphash) (struct hashtable_struct *hash_table, void *key); /* reference to the function that computes the hash value */
} HASHTABLE;

#else

typedef struct hashtable_struct /* hash table descriptor */
{
    size_t hash_size; /* size of hash table */
    HASH *table; /* reference to hash table */
    HASH *(*mkhash) (); /* reference to the function that allocates data and hash elements */
    int (*cmphash) (); /* reference to the function that compares element's keys */
    void (*rmhash) (); /* reference to the function that deallocates data hash elements */
    int (*comphash) (); /* reference to the function that computes the hash value */
} HASHTABLE;

#endif

static HASH *temp_hash; /* temporary HASH reference in SWAP() */

#define SWAP(x,y) temp_hash = (x);(x) = (y);(y) = temp_hash /* linked list link element manipulator */

#ifdef __STDC__

static void handler (int sig, siginfo_t *si, void *data); /* signal handler to retreive data */
static void print_message (int retval); /* print any error messages */
static char *binarysearch (char *begin, char *end, char *string, size_t string_len, int partial_key, char *BOFmap, char *EOFmap); /* recursively binary search a memory area */
static int strtoken (char *string, char *parse_array, char **parse, char *delim); /* parse a record based on sequential delimiters */
static int hash_init (HASHTABLE *hash_table); /* initialize the hash table */
static int hash_insert (HASHTABLE *hash_table, void *key); /* insert a key into the hash table */
static HASH *hash_find (HASHTABLE *hash_table, void *key); /* find a key in the hash table */
static void hash_term (HASHTABLE *hash_table); /* remove a hash table */
static int hash_text (HASHTABLE *hash_table, void *key); /* compute the hash value for a text key */
static int cmphash (void *data, HASH *element); /* function to compare a text key with a hash table's element key */
static HASH *mkhash (void *key); /* function to allocate a text hash table element and data */
static void rmhash (HASH *element); /* function to deallocate a text hash table element and data */
static int interpret (char *buffer); /* interpret a command */

#else

static void handler (int sig, siginfo_t *si, void *data); /* signal handler to retreive data */
static void print_message (int retval); /* print any error messages */
static char *binarysearch (); /* recursively binary search a memory area */
static int strtoken (); /* parse a record based on sequential delimiters */
static int hash_init (); /* initialize the hash table */
static int hash_insert (); /* insert a key into the hash table */
static HASH *hash_find (); /* find a key in the hash table */
static void hash_term (); /* remove a hash table */
static int hash_text ();  /* compute the hash value for a text key */
static int cmphash (); /* function to compare a text key with a hash table's element key */
static HASH *mkhash (); /* function to allocate a text hash table element and data */
static void rmhash (); /* function to deallocate a text hash table element and data */
static int interpret (); /* interpret a command */

#endif

static char default_dir[] = "."; /* default directory to watch, CWD */

static HASHTABLE hash_command = {2729, (HASH *) 0, mkhash, cmphash, rmhash, hash_text}; /* declare the hash table descriptor for the command text keys*/

#define COMMAND_TOKEN_SEPARATORS " \t\n\b\r" /* command field separators */

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    char inchar, /* character from input file */
         *db_name = (char *) 0, /* database file name */
         *log_name = (char *) 0, /* default log file */
         *dir = default_dir, /* default directory to watch, CWD */
         *begin, /* begining of memory mapped file */
         *end, /* end of memory mapped file */
         inbuffer[BUFLEN + 1], /* input buffer */
         *sentinel = &inbuffer[BUFLEN], /* reference to sentinel character in input buffer */
         parsebuffer[BUFLEN], /* parsed i/o buffer */
         *token[BUFLEN / 2], /* reference to tokens in parsed i/o buffer */
         token_separators[] = TOKEN_SEPARATORS, /* database file record field separators */
         outbuffer[BUFLEN + 1], /* output buffer */
         *ref, /* character reference in input buffer */
         *record_match, /* reference to beginning of matched record */
         *ptr; /* general character reference for string operations */

    int retval = NOERROR, /* return value, assume no error */
        logfile, /* descriptor of log file to watch */
        fd, /* file/directory descriptor */
        databasefile, /* database file descriptor */
        partial_key = 1, /* partial key search, 0 = no, 1 = yes */
        field_number = 0, /* log file field number for partial key match */
        fields, /* number of fields in a database record */
        verbose = 0, /* verbose mode flag, 0 = no, 1 = yes */
        number, /* number of command interpreter keywords in hash table */
        i, /* general integer counter */
        c; /* command line switch */

    size_t filesize; /* size of database file */

    ssize_t cnt_in = 0, /* count of characters read */
            cnt_out; /* count of characters to write */

    caddr_t mapbegin; /* reference to beginning of mmap'ed database file */

    struct sigaction act; /* sigaction(2) structure */

    struct stat fileinfo; /* structure for the status of the database file */

    HASH *element;  /* reference to hash table element */

    *sentinel = '\0'; /* sentinel character in input buffer */

    while ((retval == NOERROR) && ((c = getopt (argc, argv, "d:f:l:Vv")) != EOF)) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'd': /* request for database file name? */

                db_name = optarg; /* yes, save the database file name */
                break;

            case 'f': /* request for log file field number for partial key match */

                if ((field_number = atoi (optarg)) < 0) /* save the log file field number for partial key match; must be positive or zero */
                {
                    retval = EARGS; /* assume not enough arguments */
                }

                break;

            case 'l': /* request for log file name? */

                log_name = optarg; /* yes, save the log file name */
                break;

            case 'V': /* request for verbose mode? */

                verbose = 1; /* yes, set the verbose mode flag, 0 = no, 1 = yes */
                break;

            case 'v':

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    if (retval == NOERROR) /* any errors? */
    {
        retval = EARGS; /* assume not enough arguments */

        if (log_name != (char *) 0 && db_name != (char *) 0) /* both log file and database file names supplied? */
        {
            retval = ELFILE; /* assume error opening log file to watch */

            if ((logfile = open (log_name, O_RDONLY)) != -1) /* open the log file to watch */
            {
                retval = ELSEEK; /* assume error seeking EOF of log file to watch */

                if (lseek (logfile, (off_t) 0, SEEK_END) != (off_t) -1) /* go to the end of the log file to watch */
                {
                    retval = EDSTAT; /* assume error in status of database file */

                    if (stat (db_name, &fileinfo) == 0) /* get the status of the database file */
                    {

                        if ((filesize = (size_t) fileinfo.st_size) > (size_t) 0) /* save the size of the database file */
                        {
                            retval = EDOPEN; /* assume error opening database file */

                            if ((databasefile = open (db_name, O_RDONLY)) != -1) /* open the database file */
                            {
                                retval = EHINIT; /* assume hash table already initilaized */

                                if ((retval = hash_init (&hash_command)) == NOERROR) /* initialize the command hash table */
                                {
                                    number = sizeof (keyword) / sizeof (char *); /* get the number of keywords */

                                    for (i = 0; (i < number) && (retval == NOERROR); i ++) /* for each keyword */
                                    {
                                        (void) strcpy (inbuffer, keyword[i]); /* save a copy of the keyword */

                                        if ((retval = hash_insert (&hash_command, inbuffer)) == NOERROR) /* add a keyword to the command hash table */
                                        {
                                            retval = EHKEY; /* assume hash table key not found */

                                            if ((element = hash_find (&hash_command, inbuffer)) != (HASH *) 0) /* find the keyword in the command hash table */
                                            {
                                                element-> command = i; /* add the command keyword */
                                                retval = NOERROR; /* assume no error */
                                            }

                                        }

                                    }

                                    retval = EDMAP; /* assume error mapping database file */

                                    if ((mapbegin = mmap ((void *) 0, filesize, PROT_READ, MAP_PRIVATE, databasefile, (off_t) 0)) != (caddr_t) 0) /* map the database file */
                                    {
                                        retval = EDB;  /* assume error in database file */

                                        if (mapbegin[filesize - 1] == '\n') /* verify that the last record of the memory mapped database file has a terminating '\n' character as required for binarysearch() */
                                        {
                                            begin = (char *) mapbegin; /* reference the begining of memory mapped file */
                                            end = &mapbegin[filesize - 1]; /* reference the end of memory mapped file */
                                            act.sa_sigaction = handler; /* sigaction function, handler() */
                                            retval = ESIGINT; /* assume error initializing signal handler */

                                            if (sigemptyset (&act.sa_mask) == 0) /* initialize the signal set to empty, with all signals excluded from the set */
                                            {
                                                act.sa_flags = SA_SIGINFO; /* action is to provide information */

                                                if (sigaction (SIGRTMIN + 1, &act, (struct sigaction *) 0) == 0) /* change the action of the signal */
                                                {
                                                    dir = dirname (log_name); /* get the directory name that contains the log file to watch */
                                                    retval = EPOPEN; /* assume error opening directory */

                                                    if ((fd = open (dir, O_RDONLY | O_DIRECTORY)) != 0) /* open the directory */
                                                    {
                                                        retval = EFCNTL; /* assume error manipulating directory file descriptor */

                                                        if (fcntl (fd, F_SETSIG, SIGRTMIN + 1) != -1) /* signal when I/O becomes available on the directory */
                                                        {

                                                            if (fcntl (fd, F_NOTIFY, DN_ACCESS | DN_MODIFY | DN_CREATE | DN_RENAME | DN_DELETE | DN_ATTRIB | DN_MULTISHOT) != -1) /* notify when any file in the directory changes */
                                                            {
                                                                ref = inbuffer; /* reset the character reference in input buffer */
                                                                *ref = '\0'; /* terminate the record */
                                                                retval = NOERROR; /* assume no error */

                                                                while (retval == NOERROR) /* while forever */
                                                                {
                                                                    (void) pause (); /* wait for the signal */
                                                                    retval = ELREAD; /* assume error reading log file */

                                                                    /* retval == ELREAD means no error in progress; any other error means to drop through with the error, (without reading the next record); an error in reading from the log file to watch means drop through with the read error */

                                                                    while ((retval == ELREAD) && ((cnt_in = read (logfile, &inchar, sizeof (char)) == sizeof (char)))) /* read characters from the log file to watch */
                                                                    {
                                                                        retval = EOVRFL; /* assume error in input buffer overflow */

                                                                        if (ref < sentinel) /* buffer overflow? */
                                                                        {
                                                                            *ref = inchar; /* save the character in the input buffer */
                                                                            ref ++; /* prepare for next character reference in input buffer */
                                                                            *ref = '\0'; /* terminate the record; may overwrite the sentinel '\0' */
                                                                            retval = ELREAD; /* assume error reading log file */

                                                                            if (inchar == '\n') /* character an EOL? */
                                                                            {

                                                                                if (verbose == 1) /* verbose mode flag, 0 = no, 1 = yes, set? */
                                                                                {
                                                                                    (void) printf ("Input = %s", inbuffer); /* yes, print the input buffer */
                                                                                }

                                                                                cnt_out = ref - inbuffer; /* save the count of characters read, (count does contains the '\n'; inbuffer[cnt_out] is the '\0' */

                                                                                if ((fields = strtoken (inbuffer, parsebuffer, token, token_separators)) > field_number) /* parse the database record into fields, skip the record if there are field_number, or less, many fields; removes the '\n' */
                                                                                {

                                                                                    if (verbose == 1) /* verbose mode flag, 0 = no, 1 = yes, set? */
                                                                                    {
                                                                                        (void) printf ("Token = %s\n", token[field_number]); /* yes, print the token */
                                                                                    }

                                                                                    if ((record_match = binarysearch (begin, end, token[field_number], strlen (token[field_number]), partial_key, begin, end)) != (char *) 0) /* search the database map file for the field */
                                                                                    {

                                                                                        if (verbose == 1) /* verbose mode flag, 0 = no, 1 = yes, set? */
                                                                                        {
                                                                                            (void) printf ("Key = %s\n", token[field_number]); /* yes, print the key */
                                                                                        }

                                                                                        ptr = strchr (record_match, '\n'); /* find the EOL of the record containing the field in the database file; the database file was checked for the last character = '\n' when it was opened, so it does contain a '\n' */
                                                                                        cnt_out = ptr - record_match; /* save the length of the record containing the field in the database file, (count contains the '\n'; record_match[cnt_out] is the '\n') */
                                                                                        retval = EDBOVRFL; /* assume error in database file */

                                                                                        if (cnt_out < BUFLEN) /* database file record too long? */
                                                                                        {
                                                                                            retval = ELREAD; /* assume error reading log file */
                                                                                            (void) strncpy (outbuffer, record_match, (size_t) cnt_out); /* make a copy of the record containing the field in the database file; the '\n' is not copied */
                                                                                            outbuffer[cnt_out] = '\0'; /* change the '\n' that wasn't copied to a '\0' for parsing */

                                                                                            if ((fields = strtoken (outbuffer, parsebuffer, token, token_separators)) > 1) /* parse the record containing the field in the database file into fields, skip the record if there are no fields */
                                                                                            {

                                                                                                if (verbose == 1) /* verbose mode flag, 0 = no, 1 = yes, set? */
                                                                                                {
                                                                                                    (void) printf ("Command = %s\n", token[1]); /* yes, print the command */
                                                                                                }

                                                                                                /* cnt_out = strlen (token[1]); */ /* save the length of the next token in the database file */

                                                                                                if ((retval = interpret (token[1])) == NOERROR) /* interpret the command */
                                                                                                {
                                                                                                    retval = ELREAD; /* assume error reading log file */
                                                                                                }


                                                                                            }

                                                                                        }

                                                                                    }

                                                                                }

                                                                                ref = inbuffer; /* reset the character reference in input buffer */
                                                                                *ref = '\0'; /* terminate the record */
                                                                            }

                                                                        }

                                                                    }

                                                                    retval = ((cnt_in == 0) && (retval == ELREAD)) ? NOERROR : retval; /* a count of zero, and error reading log file to watch was an EOF or ^D, a normal exit */
                                                                }

                                                            }

                                                        }

                                                        if (close (fd) != 0) /* close the directory */
                                                        {
                                                            retval = (retval == NOERROR) ? EPCLOSE : retval; /* assume error closing the directory */
                                                        }

                                                    }

                                                }

                                            }

                                        }

                                        if (munmap ((void *) mapbegin, filesize) != 0) /* unmap the database file */
                                        {
                                            retval = (retval == NOERROR) ? EDUNMAP : retval; /* assume error unmapping the database file */
                                        }

                                    }

                                    if (close (databasefile) != 0) /* close the database file */
                                    {
                                        retval = (retval == NOERROR) ? EDCLOSE : retval; /* assume error closing the database file */
                                    }

                                }

                                hash_term (&hash_command); /* terminate the command hash table */
                            }

                        }

                    }

                }

                if (close (logfile) != 0) /* close the log file to watch */
                {
                    retval = (retval == NOERROR) ? ELCLOSE : retval; /* assume error closing the log file to watch */
                }

            }

        }

    }

    print_message (retval); /* print any error messages */
    exit (retval); /* exit with the error value */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Signal handler to retreive data.

static void handler (int sig, siginfo_t *si, void *data);

I) Load the data into globally visable variables.

Returns nothing.

*/

static void handler (int sig, siginfo_t *si, void *data)
{
  event_fd = si->si_fd; /* save the file/directory descriptor */
  event_sig = sig; /* save the signal */
  event_data = data; /* save any data */
}

/*

Print any error messages.

static void print_message (int retval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer retval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer retval:

        1) If retval is zero, print nothing-normal/successful program
           exit.

        2) Else if retval is unity, print help.

        3) Else retval is not zero or unity, it is an error code, print
           the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int retval)

#else

static void print_message (retval)
int retval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (retval) /* which return value */
    {

        case 0: /* program ended without errors, print nothing */

            break;

        case 1: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[retval]);
            break;

    }

}

/*

Recursively binary search a memory area.

static char *binarysearch (char *begin, char *end, char *string, size_t string_len, int partial_key, char *BOFmap, char *EOFmap);

I) Binary search a memory area, referenced by the begin and end
variables, for string. The variable begin references the beginning of
the area to be searched; the variable end references the end of the
area to be searched, and will always reference the '\n' terminating
character of the last record of the memory area, (which is a
requirment); the variable string is the null terminated string to
search for, and has a length represented by the variable string_len.

II) Requirements:

    A) The memory area size must be greater than zero, (this is a
    requirement for mmap(1), also.)

    B) The last character of the memory area must be a '\n' text file
    record terminator, e.g., the last record, (which, minimally, may
    be the only record,) must be terminated.

    C) The records in the memory area must be in lexical sequence,
    (consistent with strcmp() and strncmp().)

III) The general strategy is:

    A) For each iteration in the recursion:

        1) Divide the memory area exactly in two.

        2) Work from this middle, in both directions, to find the end
        of the preceding record, (or beginning of file, whichever
        comes first) and, the end of the current record:

            a) The current record contains the first instance of the
            the terminating character at, or past, the middle.

            b) The current record's beginning is the next character
            after the preceding records's terminating character, or
            the beginning of the memory area.

        3) Compare this record with the string for a match.

        4) If not a match, decide which direction to move for the next
        iteration, and call this function, again.

        5) The recursive binary search terminates when a match is
        found, or when there is no more memory area to search.

    B) Return a reference to the matched area in memory, or null if no
    match.

Returns a reference to the matched string, null if no match.

*/

#ifdef __STDC__

static char *binarysearch (char *begin, char *end, char *string, size_t string_len, int partial_key, char *BOFmap, char *EOFmap)

#else

static char *binarysearch (begin, end, string, string_len, BOFmap, EOFmap)
char *begin;
char *end;
char *string;
size_t string_len;
int partial_key;
char *BOFmap;
char *EOFmap;

#endif

{
    int retval; /* return for string comparisons */

    char *middle, /* reference to the middle between begin and end */
         *BOFmiddle_record, /* reference to the beginning of the record in the middle */
         *EOFmiddle_record, /* reference to the end of the record in the middle */
         *BOFnext_record, /* beginning of next record */
         *EOFprevious_record, /* end of previous record */
         *match = (char *) 0; /* reference to the beginning of the record in the middle if a match is found; null if no match */

    size_t middle_record_len; /* length record in the middle */

/* #define QA */ /* comment out to disable QA/Debuging */

#ifdef QA

    /*

    The Unix dictionary file, "sort -u /usr/local/lib/dict/websters >
    testfile", is a good QA test candidate.

    Also "thisprogram testfile < testfile > out" without QA #defined,
    and "cmp testfile out" is a good QA test candidate, (to verify
    that all words are successfully found,) where thisprogram is the
    name of the program containing this function.

    Test for the pathological cases, too; duplicate records, null
    records, single null records, etc.

    */

    char QAbuf[BUFSIZ], /* QA print buffer */
         *QAbegin, /* beginning of record to be printed */
         *QAend; /* end of record to be printed */

    size_t QAreclen; /* QA record length in print buffer */

#endif

    middle = BOFmiddle_record = EOFmiddle_record = begin + ((end - begin) / 2); /* reference the middle between begin and end */

    while ((BOFmap < BOFmiddle_record) && (*(BOFmiddle_record - 1) != '\n')) /* step backwards to the beginning of the record in the middle */
    {
        BOFmiddle_record --;
    }

    while ((EOFmiddle_record < EOFmap) && (*EOFmiddle_record != '\n')) /* step forwards to the end of the record in the middle */
    {
        EOFmiddle_record ++;
    }

    begin = (BOFmiddle_record < begin) ? BOFmiddle_record : begin; /* beginning of record in the middle outside of begin? */
    end = (end < EOFmiddle_record) ? EOFmiddle_record : end; /* end of record in the middle outside of end? */
    BOFnext_record = (EOFmiddle_record < EOFmap) ? EOFmiddle_record + 1 : EOFmap; /* save the beginning of next record */
    EOFprevious_record = (BOFmap < BOFmiddle_record) ? BOFmiddle_record - 1 : BOFmap; /* end of previous record */
    middle_record_len = (size_t) (EOFmiddle_record - BOFmiddle_record); /* save the length of the record in the middle, minus the terminating '\n' */
    retval = strncmp (string, BOFmiddle_record, string_len < middle_record_len ? string_len : middle_record_len); /* compare the record in the middle with the search string */

#ifdef QA

    (void) printf ("%s, ", string); /* print the string being searched for */

    QAreclen = strchr (BOFmiddle_record, '\n') - BOFmiddle_record; /* save the length of the record being searched */
    (void) strncpy (QAbuf, BOFmiddle_record, QAreclen); /* copy the record being searched to the print buffer */
    QAbuf[QAreclen] = '\0'; /* terminate the copy of the record being searched in the print buffer */
    (void) printf ("%s, ", QAbuf); /* print the copy of the record being searched */

    QAreclen = strchr (begin, '\n') - begin; /* length of first record in begin/end */
    (void) strncpy (QAbuf, begin, QAreclen); /* copy first record in begin/end to QA print buffer */
    QAbuf[QAreclen] = '\0'; /* terminate the copy of the first record in begin/end */
    (void) printf ("(%s - ", QAbuf); /* print the copy of the first record in begin/end */

    QAend = QAbegin = end; /* end of last record in begin/end */

    while ((BOFmap < QAbegin) && (*(QAbegin - 1) != '\n')) /* step backwards to the beginning of the last record in begin/end */
    {
        QAbegin --;
    }

    (void) strncpy (QAbuf, QAbegin, (size_t) (QAend - QAbegin)); /* copy last record in begin/end to QA print buffer */
    QAbuf[QAend - QAbegin] = '\0'; /* terminate the copy of the last record in begin/end */
    (void) printf ("%s)\n", QAbuf); /* print the copy of the last record in begin/end */

#endif

    if (retval < 0) /* string < record? */
    {

        if (begin < BOFmiddle_record) /* another potential match between the begining of the record in the middle and begin? */
        {
            match = binarysearch (begin, EOFprevious_record, string, string_len, partial_key, BOFmap, EOFmap); /* return a reference to any potential match between the begining of the record in the middle and begin */
        }

    }

    else if (retval > 0) /* string > record? */
    {

        if (EOFmiddle_record < end) /* another potential match between the end of the record in the middle and end? */
        {
            match = binarysearch (BOFnext_record, end, string, string_len, partial_key, BOFmap, EOFmap); /* return a reference to any potential match between the end of the record in the middle and end */
        }

    }

    else /* match, string = record? */
    {

        if (string_len < middle_record_len) /* partial key match; string shorter than record? */
        {

            if (begin < BOFmiddle_record) /* another potential match between the begining of the record in the middle and begin? */
            {
                match = binarysearch (begin, EOFprevious_record, string, string_len, partial_key, BOFmap, EOFmap); /* return a reference to any potential match between the begining of the record in the middle and begin */
            }

            if (partial_key == 1) /* partial key searche? */
            {

                /*

                The beginning of this record, (and possibly its tab
                field delimiter,) matches the string, (and possibly
                its trailing tab field delimiter); this match is not
                the first unless the variable match == NULL, (i.e., no
                matches were found between begin and
                EOFprevious_record.)

                */

                if (match == (char *) 0) /* if no matchs prior to EOFprevious_record, this is the first match */
                {
                    match = BOFmiddle_record; /* return a reference to this record */
                }

            }

        }

        else if (string_len > middle_record_len) /* partial key match; string longer than record? */
        {

            if (EOFmiddle_record < end) /* another potential match between the end of the record in the middle and end? */
            {
                match = binarysearch (BOFnext_record, end, string, string_len, partial_key, BOFmap, EOFmap); /* return a reference to any potential match between the end of the record in the middle and end */
            }

        }

        else /* exact match */
        {
            match = BOFmiddle_record; /* return a reference to this record */
        }

    }

    return (match); /* return a reference to the match found between begin and end; null if no match */
}

/*

parse a record based on sequential delimiters

int strtoken (char *string, char *parse_array, char **parse, char *delim);

parse a character array, string, into an array, parse_array, using
consecutive characters from delim as field delimiters, point the
character pointers, token, to the beginning of each field, return the
number of fields parsed

*/

#ifdef __STDC__

static int strtoken (char *string, char *parse_array, char **parse, char *delim)

#else

static int strtoken (string, parse_array, parse, delim)
char *string,
*parse_array,
**parse,
*delim;

#endif

{
    int tokens = 0;

    (void) strcpy (parse_array, string); /* copy the string */

    parse[tokens] = strtok (parse_array, delim); /* get the 1st field */

    while (parse[tokens] != 0) /* get the remaining fields */
    {
        parse[++ tokens] = strtok ((char *) 0, delim);
    }

    return (tokens); /* return the number of tokens parsed */
}

/*

Initialize the hash table.

static int hash_init (HASHTABLE *hash_table);

I) If this function has not been executed for this HASHTABLE
   structure, allocate hash_table->hash_size many HASH structures for
   the hash table index, initialize each element in the hash table
   index's next and previous references to reference the element
   itself.

II) The required argument is a reference to the hash table descriptor
    structure, hash_table.

Returns any errors encountered, NOERROR on success.

*/

#ifdef __STDC__

static int hash_init (HASHTABLE *hash_table)

#else

static int hash_init (hash_table)
HASHTABLE *hash_table;

#endif

{
    int local_retval = EHINIT; /* local return value, assume hash table already initilaized */

    size_t i; /* hash table element counter */

    HASH *hash_ref; /* reference to element in hash table index */

    if (hash_table->table == (HASH *) 0) /* protect from repeated calls by testing if hash table index has been allocated */
    {
        local_retval = EALLOC; /* assume memory error */

        if ((hash_table->table = (HASH *) malloc (hash_table->hash_size * sizeof (HASH))) != (HASH *) 0) /* allocate the hash table index */
        {
            local_retval = NOERROR; /* assume no error */

            for (i = 0; i < hash_table->hash_size; i ++) /* initialize the hash table doubly, circular linked list */
            {
                hash_ref = &hash_table->table[i]; /* reference the element in the hash table index */
                hash_ref->next = hash_ref; /* references its own element */
                hash_ref->previous = hash_ref; /* references its own element */
            }

        }

    }

    return (local_retval); /* return any error */
}

/*

Insert a key into the hash table.

static int hash_insert (HASHTABLE *hash_table, void *key);

I) Allocate the key's hash, insert the key and data at the beginning
   of the doubly linked list for this hash table index element, which
   is implicitly addressed by the key's hash.

II) The required arguments are a reference to the hash table descriptor
    structure, hash_table, and a reference to the key value, data.

Returns any errors encountered, NOERROR on success.

*/

#ifdef __STDC__

static int hash_insert (HASHTABLE *hash_table, void *key)

#else

static int hash_insert (hash_table, key)
HASHTABLE *hash_table;
void *key;

#endif

{
    int local_retval = EHMK, /* local return value, assume hash table mkhash () failure */
        hindex; /* index of element in hash table */

    HASH *element, /* reference to new element to be inserted in hash table */
         *next_element, /* reference to next element in hash's doubly, circular linked list */
         *previous_element; /* reference to indexed element in the hash table */

    hindex = (*hash_table->comphash) (hash_table, key); /* get the key's hash */
    previous_element = &hash_table->table[hindex]; /* reference the element in the hash table index */
    next_element = previous_element->next; /* start with the first element in the hash's doubly, circular linked list */

    if ((element = (*hash_table->mkhash) (key)) != (HASH *) 0) /* get the hash element */
    {
        SWAP (previous_element->next, element->next); /* the new element is first in the doubly, circular linked list */
        SWAP (next_element->previous, element->previous);
        local_retval = NOERROR; /* assume no error */
    }

    return (local_retval); /* return any error */
}

/*

Find a key in the hash table.

static HASH *hash_find (HASHTABLE *hash_table, void *key);

I) Calculate the key's hash, search for the key and data, starting at
   the beginning of the doubly linked list for this hash table index
   element, which is implicitly addressed by the key's hash.

II) The required arguments are a reference to the hash table
    descriptor structure, hash_table, and a reference to the key
    value, data.

Returns a reference to the hash table's element referencing the data,
0 if key is not found.

*/

#ifdef __STDC__

static HASH *hash_find (HASHTABLE *hash_table, void *key)

#else

static HASH *hash_find (hash_table, key)
HASHTABLE *hash_table;
void *key;

#endif

{
    int local_retval = EHKEY, /* local return value, assume hash table key not found */
        hindex; /* index of element in hash table */

    HASH *next_element, /* reference to next element in hash index's doubly, circular linked list */
         *head_element, /* reference to head element in hash index's doubly, circular linked list */
         *element = (HASH *) 0; /* return value, assume error */

    hindex = (*hash_table->comphash) (hash_table, key); /* get the key */
    head_element = next_element = &hash_table->table[hindex]; /* reference the head element in hash index's doubly, circular linked list */

    while ((next_element = next_element->next) != head_element) /* while the element in the hash index's doubly, circular linked list is not the list's head */
    {

        if ((*hash_table->cmphash) (key, next_element) == 0) /* element compare with the key? */
        {
            element = next_element; /* yes, found it, return a reference to the element */
            local_retval = NOERROR; /* assume no error */
            break;
        }

    }

    return (element); /* return any error */
}

/*

Remove a hash table.

static void hash_term (HASHTABLE *hash_table);

I) For each element in the hash table's index, for each element in the
   doubly linked lest referenced by the hash table's index element,
   remove the key's HASH element.

II) The required argument is a reference to the hash table descriptor
    structure, hash_table.

There is no return value from this function.

*/

#ifdef __STDC__

static void hash_term (HASHTABLE *hash_table)

#else

static void hash_term (hash_table)
HASHTABLE *hash_table;

#endif

{
    int local_retval = NOERROR; /* local return value, assume no error */

    size_t i;

    HASH *element, /* reference to element being deleted from hash index's doubly, circular linked list */
         *next_element, /* reference to next element in hash index's doubly, circular linked list */
         *head_element; /* reference to head element in hash index's doubly, circular linked list */

    for (i = 0; i < hash_table->hash_size && local_retval == NOERROR; i ++) /* for each of the hash table's index's (while there are no errors) */
    {
        head_element = &hash_table->table[i]; /* reference the head element in hash index's doubly, circular linked list */
        next_element = head_element->next; /* start with the first element in the hash index's doubly, circular linked list */

        while (next_element != head_element) /* while the element in the hash's doubly, circular linked list is not the list's head */
        {
            element = next_element; /* reference the element to be deleted from the doubly, circular linked list */
            next_element = next_element->next; /* reference next element in the hash's doubly, circular linked list */
            SWAP (next_element->previous,element->previous); /* delete the element from the hash's doubly, circular linked list */
            SWAP (head_element->next,element->next);
            (*hash_table->rmhash) (element); /* delete the element, and data referenced by the element */
        }

    }

    free (hash_table->table); /* no, free the hash table index */
}

/*

Compute the hash value for a text key.

static int hash_text (HASHTABLE *hash_table, void *key);

I) The routine requires a long to be 32 bits, which represents some
   portability issues.

II) The required arguments are a reference to the hash table
    descriptor structure, hash_table, and a reference to the key
    value, key.

Returns the hash value for the key.

*/

#ifdef __STDC__

static int hash_text (HASHTABLE *hash_table, void *key)

#else

static int hash_text (hash_table, key)
HASHTABLE *hash_table;
void *key;

#endif

{
    char *char_ref; /* byte reference into data */

    unsigned long num,
                  hash_num = (unsigned long) 0; /* hash number is initially zero */

    for (char_ref = key;*char_ref != (char) 0;char_ref ++) /* for each character in the key */
    {
        hash_num = (hash_num << 4) + (unsigned long) (((int) (*char_ref) < (int) 0) ? (long) (-(*char_ref)) : (long) (*char_ref)); /* multiply the hash by 16, and add the absolute value of the character */

        if ((num = hash_num & 0xf0000000) != (unsigned long) 0) /* any high order bits in bit positions 24 - 28 ? */
        {
            hash_num = hash_num ^ (num >> 24); /* yes, divide by 16777216, and exor on itself */
            hash_num = hash_num ^ num; /* reset the high order bits to 0 */
        }

    }

    return ((int) (hash_num % (unsigned long) hash_table->hash_size)); /* return the hash number */
}

/*

Function that compares a text key with a hash table element key.

static int cmphash (void *data, HASH *element);

(note that the data reference in HASH is a void reference, and
requires a cast to the appropriate data type)

returns 0 if data and the key associated with the element key are
equal

returns non-zero if data and the key associated with the element are
not equal

*/

#ifdef __STDC__

static int cmphash (void *data, HASH *element)

#else

static int cmphash (data, element)
void *data;
HASH *element;

#endif

{
    return (strcmp (data, ((char *) element->hash_key))); /* return the value of the comparison */
}

/*

Function that allocates and loads a hash table element into a hash table.

static HASH *mkhash (void *key);

returns a reference to the element constructed if successful

returns NULL if no element was constructed

Note: local_retval, is set to EALLOC if a memory allocation error
occured

*/

#ifdef __STDC__

static HASH *mkhash (void *key)

#else

static HASH *mkhash (key)
void *key;

#endif

{
    int local_retval = EALLOC; /* local return value, assume error allocating memory */

    void *key_ref; /* reference to key */

    HASH *element = (HASH *) 0; /* reference to the hash table element */

    if ((key_ref = (char *) malloc ((1 + strlen (key)) * sizeof (char))) != (char *) 0) /* allocate area for the key */
    {

        if ((element = (HASH *) malloc (sizeof (HASH))) != (HASH *) 0) /* allocate the hash table element */
        {
            element->next = element; /* all of the elements link references reference the element itself */
            element->previous = element;
            element->hash_key = key_ref; /* reference to element's key */
            element->command = (int) NONE; /* default command is NONE */
            (void) strcpy (key_ref, key); /* save the hash table element's key as its data */
            local_retval = NOERROR; /* assume no error */
        }

    }

    return (element); /* return a reference to the hash table element */
}

/*

Function to deallocate a text hash table element.

static void rmhash (HASH *element);

returns nothing

*/

#ifdef __STDC__

static void rmhash (HASH *element)

#else

static void rmhash (element)
HASH *element;

#endif
{
    free (element->hash_key); /* free the key's area allocated in mkhash () */
    free (element); /* free the hash table element allocated in mkhash () */
}

/*

Interpret a command.

int interpret (char *buffer);

I) Data structure:

    A) The command string, buffer, containing the single command
       should be an EOS = '\0' terminated string of characters-white
       space separates the tokens, (named "fields" in the comments.)

II) Execution:

    A) There are two types of commands:

        1) Internal, one of:

            a) enable

            b) disable

        2) A system command to be executed.

    B) The commands are installed in a single hash lookup table
       initialised in main ().

Returns any errors encountered, NOERROR on success.

*/

#ifdef __STDC__

int interpret (char *buffer)

#else

int interpret (buffer)
char *buffer;

#endif

{
    static char parsebuffer[BUFLEN], /* parsed i/o buffer */
                *token[BUFLEN / 2], /* reference to tokens in parsed i/o buffer */
                command_token_separators[] = COMMAND_TOKEN_SEPARATORS; /* command field separators */

    static int enable = 1; /* enable system commands flag, 0 = no, 1 = yes */

    int retval = NOERROR, /* return value, assume no error */
        fields; /* number of fields in a record */

    HASH *element;  /* reference to hash table element */

    if ((fields = strtoken (buffer, parsebuffer, token, command_token_separators)) > 0) /* parse the record into fields, skip blank records */
    {

        if (token[0][0] != '#') /* if the first character of the first field is a '#' character, skip it */
        {

            if ((element = hash_find (&hash_command, token[0])) != (HASH *) 0) /* find the key in the command hash table */
            {

                switch (element->command) /* which command */
                {

                    case ENABLE:

                        enable = 1; /* set the enable system commands flag, 0 = no, 1 = yes */
                        break;

                    case DISABLE:

                        enable = 0; /* reset the enable system commands flag, 0 = no, 1 = yes */
                        break;

                    case NONE:

                        break;

                    default: /* not a recognized command */

                        retval = ECMD; /* assume error command */
                        break;

                }

            }

            else /* not a recognized command, must be a system call */
            {

                if (enable == 1) /* enable system commands flag, 0 = no, 1 = yes, set? */
                {
                    retval = ESYS; /* assume error system call */

                    if (system (buffer) == 0) /* execute the system command */
                    {
                        retval = NOERROR; /* assume no errors */
                    }

                }

            }

        }

    }

    return (retval); /* return any errors */
}
